/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// pdplogs.cc
#include <pdp/pdplog.h>
#include <pdp/process.h>
#include <pdp/sched_proc.h>
#include <ta/ta_group_iv.h>
#include <ta_misc/datatable_iv.h>
#include <ta_misc/datagraph.h>
#include <values.h>
#include <pdplog_iv.h>

////////////////////////////////////////////////////////////////////////////////
// important notes on data_range, log_lines, and view_range:
// data_range reflects log_lines if logging to file, otherwise it is in lines
//   	in the data object
// data_range.min is >= 0, but data_range.max is not the _size_ of the max
// 	but the _line number_ of the max:  thus it is typically log_lines - 1
//	or data.MaxLength() - 1
// view_range is in the same units as data_range
// to translate either view or data_range into indicies in the data table (data)
// 	it is necessary to subtract data_range.min from the view_range
// 	to take into account the possiblity that the data in the table is only
// 	a portion of the data in the actual log file.
////////////////////////////////////////////////////////////////////////////////


//////////////////////////
// 	PDPLog		//
//////////////////////////
 
void PDPLog::Initialize() {
  log_file.cmp = false;
  log_file.mode = taFile::NO_AUTO;
  log_lines = 0;

  data.el_typ = &TA_float_Data;
  data_bufsz = 500;		// default bufsz is 500 lines
  data_shift = .20f;		// shift 20 percent (100 lines)
  record_proc_name = false;	// don't by default..
  data_range.min = 0;
  data_range.max = -1;
  cur_proc = NULL;
  views.SetBaseType(&TA_LogView);
}

void PDPLog::InitLinks() {
  taBase::Own(log_proc, this);
  taBase::Own(log_file, this);
  taBase::Own(data, this);
  taBase::Own(data_range, this);
  taBase::Own(log_data, this);

  WinMgr::InitLinks();
  InitFile();
  log_file.Close();
}

void PDPLog::Destroy() {
  CutLinks();
}

void PDPLog::CutLinks() {
  cur_proc = NULL;
  RemoveAllUpdaters();
  WinMgr::CutLinks();
}

void PDPLog::Copy_(const PDPLog& cp) {
  log_file = cp.log_file;
  log_lines = cp.log_lines;
  log_data = cp.log_data;

  data = cp.data; 
  data_bufsz = cp.data_bufsz;
  data_shift = cp.data_shift;
  data_range = cp.data_range;
  record_proc_name = cp.record_proc_name;

  log_proc.BorrowUnique(cp.log_proc);	// this is a link group..
  cur_proc = cp.cur_proc;
  InitAllViews();
}

void PDPLog::AddUpdater(SchedProcess* sp) {
  if(sp == NULL) return;
  bool was_linked = false;
  was_linked = sp->logs.LinkUnique(this);	// add this to the list on the process
  was_linked = (log_proc.LinkUnique(sp) || was_linked); // add the process to the list on this;
  if(was_linked) {
    name = sp->name + "_" + GetTypeDef()->name;
    SyncLogViewUpdaters();
    GetHeaders();
  }
}  

void PDPLog::RemoveUpdater(SchedProcess* sp) {
  if(sp == NULL) return;
  sp->logs.RemoveLeaf(this);	// remove this to the list on the process
  log_proc.RemoveLeaf(sp);     	// remove the process to the list on this;
  SyncLogViewUpdaters();
  GetHeaders();
}  

void PDPLog::RemoveAllUpdaters() {
  int i;
  for(i=0; i<log_proc.size; i++) {
    ((SchedProcess*)log_proc.FastEl(i))->logs.RemoveLeaf(this);
  }
  log_proc.RemoveAll();
  SyncLogViewUpdaters();
}

void PDPLog::SyncLogViewUpdaters() {
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i) {
    lv->updaters.Reset();
    lv->updaters.BorrowUnique(log_proc);
  }
}

int PDPLog::Dump_Save_Value(ostream& strm, TAPtr par, int indent) {
  if(log_file.IsOpen())
    log_file.mode = taFile::APPEND; // auto-open in append mode..
  else
    log_file.mode = taFile::NO_AUTO;
  return WinMgr::Dump_Save_Value(strm, par, indent);
}

void PDPLog::UpdateAfterEdit() {
  WinMgr::UpdateAfterEdit();
  SyncLogViewUpdaters();
  LogView* lv = (LogView*)views[0];
  if(lv != NULL)
    lv->NotifyAllUpdaters();	// make sure logs are getting it from  us..
  if(log_proc.size > 1)		// always record proc name if more than one..
    record_proc_name = true;
  if(log_file.gf != NULL) {
    log_file.AutoOpen();
    log_file.mode = taFile::NO_AUTO; // prevent this from happening again..
  }
  if(log_file.IsOpen()) {
    if(log_lines == 0) {
      GetFileLength();
      if(log_file.gf->ostrm != NULL) {
	HeadToFile();
      }
    }
  }
  else {
    log_lines = 0;
    data_range.max = data.MaxLength()-1;
    data_range.min = 0;
  }
  if((log_proc.size > 0) && (name.empty() || name.contains(GetTypeDef()->name))) {
    name = ((SchedProcess*)log_proc[0])->name + "_" + GetTypeDef()->name;
  }
}

void PDPLog::InitFile() {
  if(log_file.gf == NULL) {
    if(!taMisc::iv_active)
      log_file.GetGetFile();
    else {
      log_file.gf = new taivGetFile(".","*log*",NULL,wkit->style(),false);
      taRefN::Ref(log_file.gf);
    }
  }
  log_file.fname = "";
  log_file.UpdateGF();
}

void PDPLog::SetSaveFile(const char* nm, bool no_dlg) {
  log_file.SaveAsFile(nm, no_dlg);
  GetFileLength();
  HeadToFile();
  if(taMisc::iv_active)
    SetWinName();
}

void PDPLog::SetAppendFile(const char* nm, bool no_dlg) {
  log_file.AppendFile(nm, no_dlg);
  GetFileLength();
  HeadToFile();
  if(taMisc::iv_active)
    SetWinName();
}

void PDPLog::LoadFile(const char* nm, bool no_dlg) {
  log_file.OpenFile(nm, no_dlg);
  GetFileLength();
  ReadNewLogFile();
  InitAllViews();
  if(taMisc::iv_active)
    SetWinName();
}

void PDPLog::CloseFile() {
  log_file.CloseFile();
  log_lines = 0;
  if(taMisc::iv_active)
    SetWinName();
}


//////////////////////////
//	Headers 	//
//////////////////////////

void PDPLog::NewHead(LogData& ld, SchedProcess* sproc) {
  // null is a sign to regenerate regardless
  if((sproc != NULL) && (cur_proc == sproc) && (ld.items.size == data.leaves))
    return; 

  if(sproc != NULL) {		// only send to log if its real
    cur_proc = sproc;
    HeadToLogFile(ld);
  }
  
  HeadToBuffer(ld);
  LogDataFromBuffer();		// get the log data corresponding to the buffer
  UpdateViewHeaders();		// tells views that header is updated
}

void PDPLog::UpdateViewHeaders() {
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i) {
    lv->NewHead();
  }
}

void PDPLog::SetFieldHead
(LogData& ld, int fld, int top_cnt, int subgp_n, int subgp_max, int subgp_cnt) {
  DataTable* cdt = &data;
  if(subgp_max > 0) {
    if(data.gp.size <= subgp_n) {
      cdt = (DataTable*)data.NewGp(1);
    }
    else
      cdt = (DataTable*)data.gp[subgp_n];
    top_cnt = subgp_cnt;	// use top_cnt as index into cdt group
    while(cdt->size > subgp_max) { // keep it the same size..
      cdt->Remove(cdt->size-1);
    }
  }

  if(cdt->size <= top_cnt) {
    top_cnt = cdt->size;		// just in case
    DataArray_impl* nda; 
    if(ld.IsString(fld))		// indicates a string field
      nda = cdt->NewEl(1,&TA_String_Data);
    else
      nda = cdt->NewEl(1,&TA_float_Data);
  }
  if(ld.IsString(fld)) { // this checks to make sure the array is of the
			 // correct type
    if(!cdt->FastEl(top_cnt)->InheritsFrom(&TA_String_Data)) {
      DataArray_impl* nda = new String_Data;
      cdt->Replace(top_cnt, nda);
    }
  }
  else if(!cdt->FastEl(top_cnt)->InheritsFrom(&TA_float_Data)) {
    // this checks to make sure the array is of the correct type
    DataArray_impl* nda = new float_Data;
    cdt->Replace(top_cnt, nda);
  }

  DataArray_impl* da = cdt->FastEl(top_cnt);
  if(da->AR() == NULL) {
    da->NewAR();
    da->AR()->Alloc(data_bufsz);
  }
  da->name = ld.items.FastEl(fld)->name;
  da->disp_opts = ld.items.FastEl(fld)->disp_opts;
  if(cdt->name.empty()) // this gives a reasonable name to at least monitorstats
    cdt->name = da->name.after(".");
}

void PDPLog::HeadToBuffer(LogData& ld) {
  int subgp_n = -1;		// index of current sub-group
  int subgp_max = 0;		// max number in current sub-group
  int subgp_cnt = -1;		// number in current sub-group
  int top_cnt = 0;		// number in top-level group
  int i;
  for(i=0; i< ld.items.size; i++) {
    if(ld.IsVec(i)) {
      subgp_n++;
      subgp_max = ld.GetVecN(i);
      subgp_cnt = 0;
    }
    SetFieldHead(ld, i, top_cnt, subgp_n, subgp_max, subgp_cnt);
    if((subgp_max > 0) && (subgp_cnt < (subgp_max - 1)))
      subgp_cnt++;
    else {
      if(subgp_max > 0)
	subgp_max = 0;
      else
	top_cnt++;		// if just a gp, not a new item..
    }
  }
  while(data.size > top_cnt) { // keep it the same size..
    data.Remove(data.size-1);
  }
}

void PDPLog::HeadToLogFile(LogData& ld) {
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;
  
  ostream& strm = *(log_file.gf->ostrm);

  if(record_proc_name && (cur_proc != NULL))
    strm << cur_proc->name;
  strm << "_H:\t";
  int i;
  for(i=0; i < ld.items.size; i++) {
    DataItem* it = ld.items.FastEl(i);
    if(it->HasDispOption(" NARROW, "))
      LogColumn(strm, it->name, 1); // indicates one column
    else
      LogColumn(strm, it->name, 2); // default is 2 columns
  }
  strm << "\n";
  strm.flush();
}

DataItem* PDPLog::DataItemFromDataArray(DataArray_impl* da) {
  DataItem* it = new DataItem();
  it->name = da->name;
  it->disp_opts = da->disp_opts;
  if(da->InheritsFrom(TA_String_Data))
    it->is_string = true;
  log_data.items.Add(it);
  return it;
}

void PDPLog::LogDataFromDataTable(DataTable* dt, int st_idx) {
  int i;
  for(i=st_idx; i < dt->size; i++) {
    DataArray_impl* da = dt->FastEl(i);
    DataItemFromDataArray(da);
  }
  if(dt->gp.size > 0) {
    for(i=0; i<dt->gp.size; i++) {
      DataTable* ndt = (DataTable*)dt->Gp(i);
      if(ndt->size > 0) {
	DataArray_impl* da = ndt->FastEl(0);
	DataItem* it = DataItemFromDataArray(da);
	it->vec_n = ndt->size;
	LogDataFromDataTable(ndt, 1); // start from 1 since 0 was already done..
      }
    }
  }
}

void PDPLog::LogDataFromBuffer() {
  log_data.Reset();
  LogDataFromDataTable(&data);
  log_data.InitBlankData();
}

//////////////////////////
//	  Data 		//
//////////////////////////


void PDPLog::NewData(LogData& ld, SchedProcess* sproc) {
  NewHead(ld, sproc);		// see if header info is new..

  // buffer always reflects latest contents (views can scroll within)
  if(log_file.IsOpen() && (data_range.max < log_lines -1))
    Buffer_FF();

  DataToBuffer(ld);

  if(sproc != NULL)
    DataToLogFile(ld);

  int max_lns = data.MaxLength() -1; // get new size
  if(max_lns < data_range.Range()) { // must have shifted buffers..
    if(!log_file.IsOpen()) {
      data_range.min = 0;
      data_range.max = max_lns;
    }
    else {
      data_range.max++;
      data_range.min = data_range.max - max_lns;
      data_range.MinGT(0);
    }
    if(taMisc::iv_active) {
      LogView* lv;
      taLeafItr i;
      FOR_ITR_EL(LogView, lv, views., i) {
	lv->View_FF();		// catch the views up with the new range
      }
    }
  }
  else {
    data_range.max++;
    LogView* lv;
    taLeafItr i;
    FOR_ITR_EL(LogView, lv, views., i) {
      lv->NewData();		// just update new line
    }
  }
}


void PDPLog::SetFieldData
(LogData& ld, int fld, int top_cnt, int subgp_n, int subgp_max, int subgp_cnt) {
  DataTable* cdt = &data;
  if(subgp_max > 0) {
    if(data.gp.size <= subgp_n)	return;
    cdt = (DataTable*)data.gp[subgp_n];
    top_cnt = subgp_cnt;	// use top_cnt as index into cdt group
  }

  if(cdt->size <= top_cnt) return;

  DataArray_impl* da = cdt->FastEl(top_cnt);
  if(da->AR() == NULL)
    da->NewAR();
  if(da->AR()->size >= data_bufsz) { // adding will 'overflow' buffer
    da->AR()->ShiftLeftPct(data_shift);
  }

  if(ld.IsString(fld)) {
    if(da->InheritsFrom(&TA_String_Data))
      ((String_Array*)da->AR())->Add(ld.GetString(fld));
  }
  else {
    if(da->InheritsFrom(&TA_float_Data))
      ((float_Array*)da->AR())->Add(ld.GetFloat(fld));
  }
}

void PDPLog::DataToBuffer(LogData& ld) {
  int subgp_n = -1;		// index of current sub-group
  int subgp_max = 0;		// max number in current sub-group
  int subgp_cnt = -1;		// number in current sub-group
  int top_cnt = 0;		// number in top-level group
  int i;
  for(i=0; i< ld.items.size; i++) {
    if(ld.IsVec(i)) {
      subgp_n++;
      subgp_max = ld.GetVecN(i);
      subgp_cnt = 0;
    }
    SetFieldData(ld, i, top_cnt, subgp_n, subgp_max, subgp_cnt);
    if((subgp_max > 0) && (subgp_cnt < (subgp_max - 1)))
      subgp_cnt++;
    else {
      if(subgp_max > 0)
	subgp_max = 0;
      else
	top_cnt++;		// if just a gp, not a new item..
    }
  }
}

void PDPLog::DataToLogFile(LogData&) {
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;

  ostream& strm = *(log_file.gf->ostrm);

  if(record_proc_name && (cur_proc != NULL))
    strm << cur_proc->name;
  strm << "_D:\t";

  taLeafItr i;
  DataArray_impl* da;
  FOR_ITR_EL(DataArray_impl, da, data., i) {
    String el;
    taArray_base* ar = da->AR();
    if((ar != NULL) && (ar->size > 0))
      el = ar->El_GetStr_(ar->FastEl_(ar->size-1));
    else
      el = "n/a";

    if(da->HasDispOption(" NARROW, "))
      LogColumn(strm, el, 1);
    else
      LogColumn(strm, el, 2);
  }
  strm << "\n";
  strm.flush();
  log_lines++;
}

void PDPLog::GetHeaders() {
//  if(log_proc.leaves > 0) {
    data.Reset();		// clear out old data...
//  }

  LogView* lv;
  taLeafItr vi;
  FOR_ITR_EL(LogView, lv, views., vi) {
    if(lv->viewspec != NULL)
      lv->viewspec->Reset();	// clear all viewspecs too
    lv->view_range.min =0; lv->view_range.max = -1;
    lv->InitDisplay();
  }

  taLeafItr i;
  SchedProcess* sproc;
  FOR_ITR_EL(SchedProcess, sproc, log_proc., i) {
    sproc->GenLogData();
    NewHead(sproc->log_data, NULL); // this is a "fake" head, not from a log directly
    HeadToFile();
  }
}

ostream& PDPLog::LogColumn(ostream& strm, String& str, int tabs) {
  strm << str;
  if(tabs == 2) {		// only 2 options (1 or 2)
    if(str.length() < 8)
      strm << "\t\t";
    else if(str.length() < 16)
      strm << "\t";
    else
      strm << "\t";		// always tab delimited..
  }
  else {
    if(str.length() < 8)
      strm << "\t";
    else
      strm << "\t";
  }
  return strm;
}
  
 

//////////////////////////
// 	Buffers		//
//////////////////////////
 
// checks for comments and headers,
// 0 = comment, 1 = header, 2 = data

int PDPLog::LogLineType(char* lnc) {
  String ln = String(lnc);

  // skip leading spaces if any
  while(ln.firstchar() == ' ') {
    ln = ln.after(' ');
  }
  if(ln.empty())  return 0;

  if((ln.firstchar() == '#') || (ln.before(2) == "//")) {
    return 0;
  }
  
  String id = ln.before('\t');
  if(id .empty()) return 0;

  String id_sufx = id.after('_', -1); // find after last underbar
  if(id_sufx == "H:") return 1;
  if(id_sufx == "D:") return 2;
  return 2;			// assume its data
}

// reads the file up to just before the dataline, strm must be just opened
// returns the maximum dataline number it could read

int PDPLog::FileScanTo(istream& strm, int log_ln) {
  if(log_ln <= 0) return 0;	// ready to read line 0

  int count = 0;
  String templine;
  
  while(readline(strm,templine)) {
    if(LogLineType(templine) != 2) // only count data lines
      continue;
    count++;
    if (count >= log_ln) break;
  }
  return count;		// the number of the line were ready to read
}  


int PDPLog::GetFileLength() {
  if(!log_file.IsOpen()) return 0;

  taivGetFile gf;		// use a getfile for compressed reads..
  gf.fname = log_file.fname;
  istream* strm = gf.open_read();
  if((strm == NULL) || strm->bad())
    return 0;

  int act_lines = FileScanTo(*strm, MAXINT); // read as far as we can..
  gf.Close();

  log_lines = act_lines;
  data_range.MaxLT(log_lines -1);
  return act_lines;
}

void PDPLog::HeadFromLogFile(String& hdln) { // argument is mutable..
  int itm_cnt = 0;
  if(hdln.lastchar() != '\t') hdln += "\t";
  while(hdln.contains('\t')) {
    String f = hdln.before('\t');
    if((!f.empty()) && !f.contains("_H:")) {
      DataItem* it;
      if(itm_cnt >= log_data.items.size) {
	it = new DataItem();
	log_data.items.Add(it);
      }
      else
	it = log_data.items.FastEl(itm_cnt);
      if(f != it->name) {	// names don't match
	it->Initialize();	// reset info on 
	it->name = f;
	if(f[0] == '$') {
	  it->is_string = true;
	  if(f[1] == '<') {
	    String vec_n = f.after('<');
	    vec_n = vec_n.before('>');
	    it->vec_n = (int)vec_n;
	  }
	}
	else if(f[0] == '|') {
	  it->AddDispOption("NARROW, ");
	}
	else if(f[0] == '<') {
	  String vec_n = f.after('<');
	  vec_n = vec_n.before('>');
	  it->vec_n = (int)vec_n;
	}
      }
      itm_cnt++;
    }
    hdln=hdln.after('\t');
  }
  if(log_data.index.size != log_data.items.size)
    log_data.InitBlankData();
  NewHead(log_data, NULL);
}


void PDPLog::DataFromLogFile(String& dtln) {
  // log data heads are assumed to be current..
  int cnt = 0;
  if(dtln.lastchar() != '\t') dtln += "\t";
  if(log_data.index.size != log_data.items.size) { // make sure these line up..
    log_data.InitBlankData();			   // if not, reset data
  }
  while(dtln.contains('\t') && (cnt < log_data.items.size)) {
    String f = dtln.before('\t');
    if((!f.empty()) && !f.contains("_D:")) {
      if(log_data.IsString(cnt))
	log_data.GetString(cnt) = f; // these don't existe..
      else
	log_data.GetFloat(cnt) = atof(f);
      cnt++;
    }
    dtln=dtln.after('\t');
  }
  DataToBuffer(log_data);
}


void PDPLog::LogFileToBuffer() {
  if(!log_file.IsOpen()) return;

  data.ResetData();		// clear out the arrays for new data
  data_range.MinGT(0);		// min must be greater than 0

  taivGetFile gf;		// use a getfile for compressed reads..
  gf.fname = log_file.fname;
  istream* strm = gf.open_read();

  if((strm == NULL) || strm->bad())
    return;

  int act_lines = FileScanTo(*strm, (int)data_range.min);
  if(act_lines != data_range.min) {
    log_lines = act_lines;	// we adjust to the actual size of the file..
    data_range.MaxLT(log_lines-1);
    data_range.min = data_range.max - data_bufsz + 1;
    data_range.MinGT(0);
    if(log_lines == 0) {
      gf.Close();
      return;
    }
  }

  int lno = (int)data_range.min;
  String templn;
  int rval;
  while((rval = readline(*strm,templn)) && (lno <= data_range.max)) {
    int ltyp = LogLineType(templn);
    if(ltyp == 0)
      continue;
    if(ltyp == 1) {
      HeadFromLogFile(templn);
    }
    if(ltyp == 2) {
      DataFromLogFile(templn);
      lno++;
    }
  }
  if(rval == 0) {
    log_lines = lno;
    data_range.MaxLT(lno -1 );	// reset max based on what we actually got
  }
  gf.Close();
}

void PDPLog::HeadToFile() {	// based on buffer
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;

  ostream& strm = *(log_file.gf->ostrm);

  if(record_proc_name && (cur_proc != NULL))
    strm << cur_proc->name;
  strm << "_H:\t";
  int i;
  for(i=0; i < data.leaves; i++) {
    DataArray_impl* da = data.Leaf(i);
    int wdth = 2;
    if(da->HasDispOption(" NARROW, "))
      wdth = 1;
    LogColumn(strm, da->name, wdth);
  }
  strm << "\n";
  strm.flush();
}

void PDPLog::BufferToFile() {
  if(!log_file.IsOpen() || (log_file.gf->ostrm == NULL))
    return;

  ostream& strm = *(log_file.gf->ostrm);

  int max_lns = data.MaxLength()-1;
  int ln;
  for(ln=0; ln <= max_lns; ln++) {
    if(record_proc_name && (cur_proc != NULL))
      strm << cur_proc->name;
    strm << "_D:\t";
    int from_end = max_lns - ln ;
    taLeafItr i;
    DataArray_impl* da;
    FOR_ITR_EL(DataArray_impl, da, data., i) {
      String el;
      taArray_base* ar = da->AR();
      if((ar != NULL) && (ar->size > from_end) && (from_end >= 0))
	el = ar->El_GetStr_(ar->FastEl_(ar->size-1-from_end));
      else
	el = "n/a";

      if(da->HasDispOption(" NARROW, "))
	LogColumn(strm, el, 1);
      else
	LogColumn(strm, el, 2);
    }
    strm << "\n";
  }
  strm.flush();
}

void PDPLog::Buffer_F() {
  if(!log_file.IsOpen()) return;
  int shft = (int)((float)data_bufsz * data_shift);
  data_range.max += shft;
  data_range.MaxLT(log_lines - 1);
  data_range.min = data_range.max - data_bufsz + 1;
  data_range.MinGT(0);
  LogFileToBuffer();
}

void PDPLog::Buffer_FF() {
  if(!log_file.IsOpen()) return;
  data_range.max = log_lines - 1;
  data_range.min = data_range.max - data_bufsz + 1;
  data_range.MinGT(0);
  LogFileToBuffer();
}

void PDPLog::Buffer_R() {
  if(!log_file.IsOpen()) return;
  int shft = (int)((float)data_bufsz * data_shift);
  data_range.min -= shft;
  data_range.MinGT(0);
  data_range.max = data_range.min + data_bufsz - 1;
  data_range.MaxLT(log_lines - 1);
  LogFileToBuffer();
}

void PDPLog::Buffer_FR() {
  if(!log_file.IsOpen()) return;
  data_range.min = 0;
  data_range.max = data_range.min + data_bufsz -1;
  data_range.MaxLT(log_lines - 1);
  LogFileToBuffer();
}

void PDPLog::ReadNewLogFile() {
  Buffer_FF();		// go to end
  if(!taMisc::iv_active) return;
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i) {
    lv->View_FF();		// same for all the views
  }
}

void PDPLog::Clear() {
  data.ResetData();
  data_range.min = 0; data_range.max = -1;
  if(!taMisc::iv_active) return;
  LogView* lv;
  taLeafItr i;
  FOR_ITR_EL(LogView, lv, views., i)
    lv->Clear_impl();
}


void NetLog::SetNetwork(Network* net) {
  NetLogView* lv;
  taLeafItr i;
  FOR_ITR_EL(NetLogView, lv, views., i)
    lv->SetNetwork(net);
}


//////////////////////////
// 	LogView		//
//////////////////////////


void LogView::Initialize() {
  view_bufsz = 100;
  view_shift = .2f;
  view_range.min = 0;
  view_range.max = -1;
  viewspec = NULL;
  log_mgr = NULL;
  impl_ = NULL;
}

void LogView::CutLinks() {
  log_mgr = NULL;		// prevent from cutting updater links when killed
  taBase::DelPointer((TAPtr*)&viewspec);
  if(impl_ != NULL) { delete impl_; impl_ = NULL; }
  PDPView::CutLinks();
}

void LogView::InitLinks() {
  log_mgr = GET_MY_OWNER(PDPLog);
  taBase::Own(view_range, this);
  taBase::Own(viewspecs,this);

  if(log_mgr == NULL) {
    body = wkit->label("This view does not work unless it is a view of a Log");
    return;
  }

  log_mgr->SyncLogViewUpdaters(); // make sure we're all together here..
  CreateViewSpec();
  viewspec->BuildFromDataTable(&(log_mgr->data));
  PDPView::InitLinks();
  if(impl_ == NULL) impl_ = Get_IV_Impl();
}

void LogView::Copy_(const LogView& cp) {
  view_bufsz = cp.view_bufsz;
  view_shift = cp.view_shift;
  view_range = cp.view_range;
  viewspecs = cp.viewspecs;
}

void LogView::UpdateAfterEdit() {
  PDPView::UpdateAfterEdit();
  if(!taMisc::iv_active) return;
  impl_->SetToggle(display_toggle);
  InitDisplay();
}

LogView_IV_Impl* LogView::Get_IV_Impl(){ 
  return new LogView_IV_Impl(this);
}

TypeDef* LogView::DT_ViewSpecType() {
  return &TA_DT_ViewSpec;
}

TypeDef* LogView::DA_ViewSpecType() {
  return &TA_DA_ViewSpec;
}

void LogView::CreateViewSpec() {
  if(viewspecs.gp.size == 0) {
    viewspecs.NewGp(1, DT_ViewSpecType());
    taBase::DelPointer((TAPtr*)&viewspec); // this must be old if it exists
  }
  if(viewspec == NULL) {
    taBase::SetPointer((TAPtr*)&viewspec, viewspecs.gp.DefaultEl());
    viewspec->SetBaseType(DA_ViewSpecType());
  }
}

void LogView::NewHead(){
  if(log_mgr &&(viewspec != NULL))
    viewspec->BuildFromDataTable(&(log_mgr->data));
}

void LogView::Clear() {
  if(log_mgr == NULL)	return;
  log_mgr->Clear();
  Clear_impl();
}

void LogView::Clear_impl(){};

void LogView::AddNotify(TAPtr ud) {
  if(ud->InheritsFrom(TA_SchedProcess) && (log_mgr != NULL)) {
    log_mgr->AddUpdater((SchedProcess*)ud);
  }
}

void LogView::RemoveNotify(TAPtr ud) {
  if(ud->InheritsFrom(TA_SchedProcess) && (log_mgr != NULL)) {
    log_mgr->RemoveUpdater((SchedProcess*)ud);
  }
}

void LogView::SetToggle(bool value){
  display_toggle = value;
  impl_->SetToggle(value);
}

void LogView::ToggleDisplay() {
  display_toggle = !display_toggle;
  impl_->ToggleDisplay();
}

void LogView::View_F() {
  if(!taMisc::iv_active) return;
  int shft = (int)((float)view_bufsz * view_shift);
  if(shft < 1) shft = 1;
  view_range.max += shft;
  int log_max = MAX((int)log_mgr->log_lines-1, (int)log_mgr->data_range.max);
  view_range.MaxLT(log_max); // keep it less than max
  view_range.min = view_range.max - view_bufsz +1; // always keep min and max valid
  view_range.MinGT(0);
  if(view_range.max > log_mgr->data_range.max) {
    log_mgr->Buffer_F();
  }
  view_range.MaxLT(log_mgr->data_range.max); // redo now that data_range.max is final
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  UpdateFromBuffer();
}

void LogView::View_FSF() {
  if(!taMisc::iv_active) return;
  view_range.max += view_bufsz;
  int log_max = MAX((int)log_mgr->log_lines-1, (int)log_mgr->data_range.max);
  view_range.MaxLT(log_max); // keep it less than max
  view_range.min = view_range.max - view_bufsz +1; // always keep min and max valid
  view_range.MinGT(0);
  if(view_range.max > log_mgr->data_range.max) {
    log_mgr->Buffer_F();
  }
  view_range.MaxLT(log_mgr->data_range.max); // redo now that data_range.max is final
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  UpdateFromBuffer();
}

void LogView::View_FF() {
  if(!taMisc::iv_active) return;
  view_range.max = MAX(log_mgr->log_lines-1, log_mgr->data_range.max);
  view_range.min = view_range.max - view_bufsz +1; // always keep min and max valid
  view_range.MinGT(0);
  if(view_range.max > log_mgr->data_range.max) {
    log_mgr->Buffer_FF();
  }
  view_range.MaxLT(log_mgr->data_range.max); // redo now that data_range.max is final
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  UpdateFromBuffer();
}

void LogView::View_R() {
  if(!taMisc::iv_active) return;
  int shft = (int)((float)view_bufsz * view_shift);
  if(shft < 1) shft = 1;
  view_range.min -= shft;
  view_range.MinGT(0);
  view_range.max = view_range.min + view_bufsz -1; // always keep min and max valid
  int log_max = MAX((int)log_mgr->log_lines-1, (int)log_mgr->data_range.max);
  view_range.MaxLT(log_max); // keep it less than max
  if(view_range.min < log_mgr->data_range.min) {
    log_mgr->Buffer_R();
  }
  view_range.MinGT(log_mgr->data_range.min); // redo now that data_range.min is final
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max);
  UpdateFromBuffer();
}

void LogView::View_FSR() {
  if(!taMisc::iv_active) return;
  view_range.min -= view_bufsz;
  view_range.MinGT(0);
  view_range.max = view_range.min + view_bufsz -1; // always keep min and max valid
  int log_max = MAX((int)log_mgr->log_lines-1, (int)log_mgr->data_range.max);
  view_range.MaxLT(log_max); // keep it less than max
  if(view_range.min < log_mgr->data_range.min) {
    log_mgr->Buffer_R();
  }
  view_range.MinGT(log_mgr->data_range.min); // redo now that data_range.min is final
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max);
  UpdateFromBuffer();
}

void LogView::View_FR() {
  if(!taMisc::iv_active) return;
  view_range.min = 0;
  view_range.max = view_range.min + view_bufsz -1; // always keep min and max valid
  int log_max = MAX((int)log_mgr->log_lines-1, (int)log_mgr->data_range.max);
  view_range.MaxLT(log_max); // keep it less than max
  if(view_range.min < log_mgr->data_range.min) {
    log_mgr->Buffer_FR();
  }
  view_range.MinGT(log_mgr->data_range.min); // redo now that data_range.min is final
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max);
  UpdateFromBuffer();
}

void LogView::CloseWindow() {  
  if(!taMisc::iv_active || (window == NULL)) return;
  PDPView::CloseWindow();
}

void LogView::InitDisplay(){
  PDPView::InitDisplay();
}

void LogView::UpdateDisplay(TAPtr){
  PDPView::UpdateDisplay();
}

void LogView::EditViewSpec(taBase* spec){
  if(spec==NULL) return;
  spec->Edit();
}

void LogView::ToggleVisibility(taBase* spec){
  if(spec==NULL) return;
  if(spec->InheritsFrom(&TA_DA_ViewSpec)){
    DA_ViewSpec* davs = (DA_ViewSpec *) spec;
    davs->visible = !davs->visible;
    davs->UpdateAfterEdit();
    davs->UpdateView();
  }
  else if(spec->InheritsFrom(&TA_DT_ViewSpec)){
    DT_ViewSpec* dtvs = (DT_ViewSpec *) spec;
    dtvs->visible = !dtvs->visible;
    dtvs->UpdateAfterEdit();
  }
}

ivGlyph* LogView::GetPrintData() {
  return impl_->GetPrintData();
}

//////////////////////////
// 	TextLogView	//
//////////////////////////
 
void TextLogView::Initialize() {
  view_bufsz = 10;
  view_shift = .5f;
}

LogView_IV_Impl* TextLogView::Get_IV_Impl(){ 
  return new TextLogView_IV_Impl(this);
}

void TextLogView::CloseWindow() {
  if(!taMisc::iv_active || (window == NULL)) return;
  ivResource::unref(body);	body = NULL;
  LogView::CloseWindow();
}

void TextLogView::UpdateAfterEdit() {
  LogView::UpdateAfterEdit();
  UpdateDisplay();
}

TypeDef* TextLogView::DT_ViewSpecType() {
  return &TA_DT_ViewSpec;
}

TypeDef* TextLogView::DA_ViewSpecType() {
  return &TA_DA_TextViewSpec;
}

int TextLogView::ShowLess(int newsize){
  if(newsize <= 0) return 0;
  view_bufsz = MAX(newsize, 1);	// never show zero!
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max); // double check
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  // this crashes because its called within a draw..
  //  FixSize();
  return 0;
}

int TextLogView::ShowMore(int newsize){
  if(newsize <= 0) return 0;
  int oldsize = (int) (view_range.max - view_range.min +1);
  view_bufsz = MAX(newsize, 1);
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max); // double check
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  if(oldsize != (view_range.max - view_range.min + 1)) {
    // this crashes because its called within a draw..
    //    FixSize();
    //    return 0;
    return 1;
  }
  else {
    return 1;
  }
}

void TextLogView::GetBodyRep() {
  if(!taMisc::iv_active) return;
  if(body != NULL) return;
  if(impl_ == NULL) impl_ = Get_IV_Impl();
  body = impl_->GetBodyRep();
  ivResource::ref(body);
};


ivCoord TextLogView::GetFieldWidth(DA_TextViewSpec* vs){
  return(vs->width * 4.0f);
}  

void TextLogView::MakeLine(int from_end){
  taLeafItr i;
  DataArray_impl* da;
  DA_TextViewSpec* vs;
  FOR_ITR_EL(DA_TextViewSpec, vs, viewspec->, i) {
    if((vs->visible == false) || (vs->data_array == NULL)) continue;
    da = vs->data_array;
    taArray_base* ar = da->AR();
    String el;
    if((ar != NULL) && (ar->size > from_end) && (from_end >= 0)) {
      el = ar->El_GetStr_(ar->FastEl_(ar->size-1-from_end));
    }
    else {
      el = "n/a";
    }
    ((TextLogView_IV_Impl *) impl_)->AddToLine(el.chars(), GetFieldWidth(vs));
  }
}

void TextLogView::InitDisplay() {
  if(!taMisc::iv_active) return;
  if(body == NULL) return;
  ((TextLogView_IV_Impl *) impl_)->StartHeaderLine();
  taLeafItr i;
  DataArray_impl* da;
  DA_TextViewSpec* vs;
  FOR_ITR_EL(DA_TextViewSpec, vs, viewspec->, i) {
    if((vs->visible == false) || (vs->data_array == NULL)) continue;
    da = vs->data_array;
    ((TextLogView_IV_Impl *) impl_)->AddToHeaderLine(vs->display_name.chars(),
						     GetFieldWidth(vs));
  }
  ((TextLogView_IV_Impl *) impl_)->EndHeaderLine();
  UpdateFromBuffer();
  UpdateDisplay();
}

void TextLogView::UpdateDisplay(TAPtr) {
  if(!taMisc::iv_active) return;
  ((TextLogView_IV_Impl *) impl_)->UpdateDisplay();
}

void TextLogView::Reset() {
  if(!taMisc::iv_active) return;
  ((TextLogView_IV_Impl *) impl_)->RemoveLine(-1);
}

void TextLogView::Clear_impl() {
  if(!taMisc::iv_active ) return;
  Reset();
  view_range.max = -1;
  view_range.min = 0; 
  InitDisplay();
}


void TextLogView::NewHead() {
  LogView::NewHead();
  if(!IsMapped() || !display_toggle) return;
  InitDisplay();
}

void TextLogView::NewData() {
  if(!IsMapped() || !display_toggle) return;

  if((view_range.max < log_mgr->data_range.max-1) && (view_range.max > 0))
    return;			// i'm not viewing the full file here..
  MakeLine();
  ((TextLogView_IV_Impl *) impl_)->AddLine();
  view_range.max++;
  while((view_range.max - view_range.min + 1) > view_bufsz) {
    view_range.min++;
    ((TextLogView_IV_Impl *) impl_)->RemoveLine(0);
  }
  UpdateDisplay();
}


void TextLogView::FixSize(){
  // this updates the data area
  if(!taMisc::iv_active || !display_toggle) return;

  view_range.WithinRange(log_mgr->data_range);
  Reset();

  // this is the index (in view_range units) of the last line of data in the buffer
  int data_offset = log_mgr->data.MaxLength() + (int)log_mgr->data_range.min - 1;

  int ln;
  for(ln = (int)view_range.min; ln <= (int)view_range.max; ln++) {
    int from_end = data_offset - ln ;
    MakeLine(from_end);
    ((TextLogView_IV_Impl *) impl_)->AddLine();
  }
}

void TextLogView::UpdateFromBuffer() {
  FixSize();
  ((TextLogView_IV_Impl *) impl_)->UpdateFromBuffer();
}

//////////////////////////
// 	PageLogView	//
//////////////////////////
 

void PageLogView::AutoLayout(){
  if(impl_ != NULL) ((PageLogView_IV_Impl *) impl_)->AutoLayout();
}

void PageLogView::ToggleEdit(){
  if(impl_ != NULL) ((PageLogView_IV_Impl *) impl_)->ToggleEdit();
}

void PageLogView::Initialize() {
  view_bufsz = 10;
  view_shift = .5f;
}

LogView_IV_Impl* PageLogView::Get_IV_Impl(){ 
  return new PageLogView_IV_Impl(this);
}

void PageLogView::CloseWindow() {
  if(!taMisc::iv_active || (window == NULL)) return;
  ivResource::unref(body);	body = NULL;
  LogView::CloseWindow();
}

void PageLogView::UpdateAfterEdit() {
  LogView::UpdateAfterEdit();
  UpdateDisplay();
}

TypeDef* PageLogView::DT_ViewSpecType() {
  return &TA_DT_ViewSpec;
}

TypeDef* PageLogView::DA_ViewSpecType() {
  return &TA_DA_PageViewSpec;
}

int PageLogView::ShowLess(int newsize){
  if(newsize <= 0) return 0;
  view_bufsz = MAX(newsize, 1);	// never show zero!
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max); // double check
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
    FixSize();
  return 0;
}

int PageLogView::ShowMore(int newsize){
  if(newsize <= 0) return 0;
  int oldsize = (int) (view_range.max - view_range.min +1);
  view_bufsz = MAX(newsize, 1);
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max); // double check
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  if(oldsize != (view_range.max - view_range.min + 1)) {
    FixSize();
    return 0;
  }
  else {
    return 1;
  }
}

void PageLogView::GetBodyRep() {
  if(!taMisc::iv_active) return;
  if(body != NULL) return;
  if(impl_ == NULL) impl_ = Get_IV_Impl();
  body = impl_->GetBodyRep();
  ivResource::ref(body);
};


void PageLogView::MakeLine(int from_end){
  taLeafItr i;
  DataArray_impl* da;
  DA_PageViewSpec* vs;
  FOR_ITR_EL(DA_PageViewSpec, vs, viewspec->, i) {
    if((vs->visible == false) || (vs->data_array == NULL)) continue;
    da = vs->data_array;
    taArray_base* ar = da->AR();
    String el;
    if((ar != NULL) && (ar->size > from_end) && (from_end >= 0)) {
      el = ar->El_GetStr_(ar->FastEl_(ar->size-1-from_end));
    }
    else {
      el = "n/a";
    }
    ((PageLogView_IV_Impl *) impl_)->AddToLine(el.chars(), vs->pos.x, vs->pos.y);
  }
}

void PageLogView::InitDisplay() {
  if(!taMisc::iv_active) return;
  if(body == NULL) return;
  ((PageLogView_IV_Impl *) impl_)->StartHeaderPage();
  taLeafItr i;
  DataArray_impl* da;
  DA_PageViewSpec* vs;
  if(viewspec->leaves == 0)
    ((PageLogView_IV_Impl *) impl_)-> AddToHeaderPage(NULL,0,0);
  FOR_ITR_EL(DA_PageViewSpec, vs, viewspec->, i) {
    if((vs->visible == false) || (vs->data_array == NULL)) continue;
    da = vs->data_array;
    ((PageLogView_IV_Impl *) impl_)->AddToHeaderPage(vs->display_name.chars(),vs->pos.x,vs->pos.y);
  }
  ((PageLogView_IV_Impl *) impl_)->EndHeaderPage();
  UpdateFromBuffer();
  UpdateDisplay();
}


void PageLogView::UpdateFromDisplay(){ // updates positions from positions in display
  taLeafItr i;
  DataArray_impl* da;
  DA_PageViewSpec* vs;
  int count = 0;
  FOR_ITR_EL(DA_PageViewSpec, vs, viewspec->, i) {
    if((vs->visible == false) || (vs->data_array == NULL)) continue;
    da = vs->data_array;
    ((PageLogView_IV_Impl *) impl_)->GetPosition(count,vs->pos.x,vs->pos.y);
    count++;
  }
  FixSize();
  UpdateDisplay();
}

void PageLogView::UpdateDisplay(TAPtr) {
  if(!taMisc::iv_active) return;
  ((PageLogView_IV_Impl *) impl_)->UpdateDisplay();
}

void PageLogView::Reset() {
  if(!taMisc::iv_active) return;
  ((PageLogView_IV_Impl *) impl_)->RemoveLine(-1);
}

void PageLogView::Clear_impl() {
  if(!taMisc::iv_active ) return;
  Reset();
  view_range.max = -1;
  view_range.min = 0; 
  InitDisplay();
}


void PageLogView::NewHead() {
  LogView::NewHead();
  if(!IsMapped() || !display_toggle) return;
  InitDisplay();
}

void PageLogView::NewData() {
  if(!IsMapped() || !display_toggle) return;

  if((view_range.max < log_mgr->data_range.max-1) && (view_range.max > 0))
    return;			// i'm not viewing the full file here..
  MakeLine();
  ((PageLogView_IV_Impl *) impl_)->AddLine();
  ((PageLogView_IV_Impl *) impl_)->vbox->modified(2); // so that the hungrybox
						      // actuall gets realloced
						      // in the updatedisplay below
  view_range.max++;
  while((view_range.max - view_range.min + 1) > view_bufsz) {
    view_range.min++;
    ((PageLogView_IV_Impl *) impl_)->RemoveLine(0);
  }
  UpdateDisplay();
}


void PageLogView::FixSize(){
  // this updates the data area
  if(!taMisc::iv_active || !display_toggle) return;

  view_range.WithinRange(log_mgr->data_range);
  Reset();

  // this is the index (in view_range units) of the last line of data in the buffer
  int data_offset = log_mgr->data.MaxLength() + (int)log_mgr->data_range.min - 1;

  int ln;
  for(ln = (int)view_range.min; ln <= (int)view_range.max; ln++) {
    int from_end = data_offset - ln ;
    MakeLine(from_end);
    ((PageLogView_IV_Impl *) impl_)->AddLine();
  }
}

void PageLogView::UpdateFromBuffer() {
  FixSize();
  ((PageLogView_IV_Impl *) impl_)->UpdateFromBuffer();
}



//////////////////
//  NetLogView  //
//////////////////

void NetLogView::Initialize(){
  view_bufsz = 1;
  view_shift = 1;
  network = NULL;
}

void NetLogView::InitLinks() {
  LogView::InitLinks();
  if(updaters.size == 0) return;
  SchedProcess* proc = (SchedProcess*)updaters.FastEl(0);
  if(proc->network != NULL)
    SetNetwork(proc->network);	// use process network by default
}

void NetLogView::CutLinks() {
  LogView::CutLinks();
  taBase::DelPointer((TAPtr*)&network);
}

TypeDef* NetLogView::DA_ViewSpecType() {
  return &TA_DA_NetViewSpec;
}

void NetLogView::UpdateAfterEdit(){
  LogView::UpdateAfterEdit();
  UpdateDisplay();
}

void NetLogView::SetNetwork(Network* net) {
  taBase::SetPointer((TAPtr*)&network, net);
  UpdateDisplay();
}

void NetLogView::NewHead(){
  LogView::NewHead();
  if(!display_toggle) return;
  InitDisplay();
}

void NetLogView::NewData(){
  if(!taMisc::iv_active || !display_toggle) return;
  UpdateDisplay();
}

void NetLogView::InitDisplay() {
  if(network == NULL) {
    if(updaters.size == 0) return;
    SchedProcess* proc = (SchedProcess*)updaters.FastEl(0);
    if(proc->network != NULL)
      SetNetwork(proc->network);	// use process network by default
  }
  else
    UpdateDisplay();
}

void NetLogView::UpdateDisplay(TAPtr) {
  if(!taMisc::iv_active || (network == NULL)) return;
  NetView* netview;
  taLeafItr j;
  FOR_ITR_EL(NetView, netview, network->views., j) {
    if(netview->display_toggle == false) continue;
    bool newlabel = false;
    taLeafItr i;
    DataArray_impl* da;
    DA_NetViewSpec* vs;
    FOR_ITR_EL(DA_NetViewSpec, vs, viewspec->, i) {
      if((vs->visible == false) || (vs->data_array == NULL)) continue;
      da = vs->data_array;
      NetViewLabel* vl = NULL;
      if ((vs->label_index < 0) ||
	  ((netview->labels.size) <= vs->label_index)) {
	vl = (NetViewLabel *) netview->labels.New(1,&TA_NetViewLabel);
	vs->label_index = netview->labels.size -1;
      }
      else  {
	vl = (NetViewLabel *) netview->labels.El(vs->label_index);
      }
      
      vl->name = vs->display_name + ":";
      taArray_base* ar = da->AR();
      String el;
      if((ar != NULL) && (ar->size > 0)){
	el = ar->El_GetStr_(ar->FastEl_(ar->size-1));
      }
      else {
	el = "n/a";
      }
      vl->name += el;
      vl->UpdateAfterEdit(); //    vl->MakeText();
    }
    if(newlabel == true) netview->InitDisplay();
    else { netview->UpdateDisplay(); }
  }
}

void NetLogView::UpdateFromBuffer(){
  UpdateAfterEdit();
}


//////////////////////////
//    GridLogView	//
//////////////////////////


LogView_IV_Impl* GridLogView::Get_IV_Impl(){ 
  return ((LogView_IV_Impl *) new GridLogView_IV_Impl(this));
}


// these are defined here so the grid_object does not have to know
// about the gridlog view, and is more general

static bool GLVGetScaleOn(void* glv, char* name){
  return (((GridLogView *) glv)->scale_on.Find(name) != -1);
}
static void GLVScriptScaleOn(void* glv, char* name, bool on){
  ((GridLogView *) glv)->ScriptScaleOn(name,on);
}
static float GLVGetScaleRange(void* glv, ScaleRange** sr) {
  return ((GridLogView*) glv)->GetScaleRange(sr);
}
static int GLVGetBlockSize(void* glv) {
  return ((GridLogView*) glv)->block_size;
}
static int GLVGetBlockBorder(void* glv) {
  return ((GridLogView*) glv)->block_border_size;
}

TypeDef* GridLogView::DT_ViewSpecType() {
  return &TA_DT_GridViewSpec;
}

TypeDef* GridLogView::DA_ViewSpecType() {
  return &TA_DA_GridViewSpec;
}


float GridLogView::GetSingleRange(float initial,int minev,int maxev,
				  DataTable* t, int n){

  int offsetmin;
  int offsetmax;

  int min_from_end = t->MaxLength() - minev;
  int max_from_end = t->MaxLength() - maxev;

  float srange = initial;
  if(t==NULL) return initial;
  int j,k;
  if(n != -1) {
    taArray_base* ar = t->El(n)->AR();
    if(ar == NULL)
      return initial;
    offsetmin = MAX(0, ar->size-min_from_end);
    offsetmax = MIN(ar->size-1, ar->size-max_from_end);
    for(j=offsetmin;j<=offsetmax;j++){
      float val =  fabs((float) *((float *)ar->FastEl_(j)));
      srange = MAX(srange,val);
    }
  }
  else {
    for(k=0;k<t->size;k++){
      float sngrng = GetSingleRange(initial,minev,maxev,t,k);
      srange = MAX(srange, sngrng);
    }
    for(k=0;k<t->gp.size;k++){
      float sngrng = GetSingleRange(srange,minev,maxev,(DataTable *) t->Gp(k),-1);
      srange = 	MAX(srange,sngrng);
    }
  }
  return srange;
}
  

float GridLogView::GetRange(float initial, int minev, int maxev) {
  float srange = initial;
  scale_on.Reset();
  int i;
  for(i=0;i<obj->headerbox->count();i++){
    HeaderLabel* hl = (HeaderLabel *) obj->headerbox->component(i);
    DataTable* t;
    DT_GridViewSpec* v;
    int n;
    if (!(hl->GetTableIndex(&t,&v,&n))) continue;
    if(n == -1) scale_on.Add(v->name);
    else scale_on.Add(((*v)[n])->name);
    float sngrng = GetSingleRange(initial,minev,maxev,t,n);
    srange = MAX(srange, sngrng);
  }
  return srange;
}


float GridLogView::GetScaleRange(ScaleRange** range) {
  if(range!=NULL) { // the user just wants the scalerange object
    *range = &scale_range;
    return 0.0f;
  }
  float srange = 1.0f;
  int minev,maxev;

  actual_range.min = view_range.min - log_mgr->data_range.min;
  actual_range.max = view_range.max - log_mgr->data_range.min;

  switch(scale_range.grain) {
  case ScaleRange::BY_ROW:
    minev = maxev = (int) actual_range.min;
    srange = GetRange(scale_range.initial,minev,maxev);
    break;
  case ScaleRange::BY_VISIBLE:
    minev = (int)  actual_range.min;	
    maxev = (int)  actual_range.max;
    srange = GetRange(scale_range.initial,minev,maxev);
    break;
  case ScaleRange::BY_BUFFER:
    minev = (int)  0; maxev = (int)  log_mgr->data.MaxLength() -1;
    srange = GetRange(scale_range.initial,minev,maxev);
    break;
  case ScaleRange::BY_FILE:
    minev = (int)  0; maxev = (int)  log_mgr->data.MaxLength() -1;
    srange = GetRange(scale_range.initial,minev,maxev);
    srange = MAX(obj->hcsb->range,srange);
    break;
  case ScaleRange::FIXED:
    srange = scale_range.initial;
    break;
  }
  return srange;
}


void GridLogView::Copy_(const GridLogView& cp) {
  fill_type = cp.fill_type;
  block_size = cp.block_size;
  block_border_size = cp.block_border_size;
  taBase::SetPointer((TAPtr*)&colorspec, cp.colorspec);
  scale_range = cp.scale_range;
  actual_range = cp.actual_range;
  scale_on = cp.scale_on;
}

void GridLogView::InitLinks(){
  taBase::Own(scale_on, this);
  log_mgr = GET_MY_OWNER(PDPLog);
  if(log_mgr == NULL) {
    LogView::InitLinks();
    return;
  }

  CreateViewSpec();
  // set top level view spec to not use gp_name
  ((DT_GridViewSpec *) viewspec)->use_gp_name = false;

  if(taMisc::iv_active) {
    obj = new DT_CEdit_Obj(&(log_mgr->data),(DT_GridViewSpec *) viewspec);
    obj->scale_ctrl_obj = this;
    obj->GetScaleRange = GLVGetScaleRange;
    obj->GetBlockSize = GLVGetBlockSize;
    obj->GetBlockBorder = GLVGetBlockBorder;
    obj->GetScaleOn = GLVGetScaleOn;
    obj->ScriptScaleOn = GLVScriptScaleOn;
  }
  taBase::Own(scale_range,this);
  LogView::InitLinks(); 
  if(taMisc::iv_active) ((GridLogView_IV_Impl *) impl_)->CreateAdjuster();
}

void GridLogView::CloseWindow() {
  ivResource::unref(body);	body = NULL;
  LogView::CloseWindow();
  if(obj != NULL) { delete obj; obj = NULL; }
}

// void GridLogView::SetScaleItem(DataTable* table, int index){
//   obj->hcsb->range = 0; // if you change the item then reset the range.
//   obj->hcsb->zero = 0; // if you change the item then reset the range.
//   // set scale item reset grain. otherwise the new table/index would
//   // be irrelevant
//   if(scale_range.grain == ScaleRange::FIXED) {
//     scale_range.grain = ScaleRange::BY_BUFFER;
//     scale_range.initial = 0;
//     scale_range.offset = 0;
//   }
// }

int GridLogView::ShowLess(int newsize){
  if(newsize < 0) return 0;
  view_bufsz = MAX(newsize, 1);	// never show zero
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max); // double check
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  FixSize();
  return 0;
}

int GridLogView::ShowMore(int newsize){
  if(newsize < 0) return 0;
  int oldsize = (int) (view_range.max - view_range.min +1);
  view_bufsz = MAX(newsize, 1);	// never show zero
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT(log_mgr->data_range.max); // double check
  view_range.min = view_range.max - view_bufsz +1;
  view_range.MinGT(log_mgr->data_range.min);
  if(oldsize != (view_range.max - view_range.min + 1)) {
    FixSize();
    return 0;
  }
  else {
    return 1;
  }
}

void GridLogView::NewHead() {
  LogView::NewHead();
  if(!IsMapped() || !display_toggle) return;
  scale_on.Reset(); // clear out scale
  if((obj != NULL) && (obj->dtgvs != NULL) &&
     (obj->dtgvs->CheckGeom() == true)) obj->dtgvs->UpdateLayout();
  InitDisplay();
}

void GridLogView::NewData(){
  if(!IsMapped() || !display_toggle ) return;
  if((view_range.max < log_mgr->data_range.max-1) && (view_range.max > 0))
    return;			// not viewing the whole range, don't update
  view_range.max++;
  view_range.MaxLT(log_mgr->data_range.max); // double check
  view_range.MinGT(view_range.max - view_bufsz + 1); 
  view_range.WithinRange(log_mgr->data_range); // triple check
  UpdateFromBuffer();
}

void GridLogView::FixSize(){
  if(!taMisc::iv_active) return;
  actual_range.min = view_range.min - log_mgr->data_range.min;
  actual_range.max = view_range.max - log_mgr->data_range.min;
  obj->FixSize(&actual_range);
//  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

void GridLogView::UpdateFromBuffer(){
  FixSize();
}

void GridLogView::Reset(){
  if(!taMisc::iv_active) return;
//  obj->Reset();
  obj->FixSize(&view_range);
  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

void GridLogView::Clear_impl(){
  if(!taMisc::iv_active) return;
  Reset();
  view_range.max = -1;
  view_range.min = 0; 
  InitDisplay();
}


int GLVMore(int* glv, int newsize, HungryBox*) {
 return ((GridLogView*) glv)->ShowMore(newsize);
}

int GLVLess(int* glv, int newsize, HungryBox*) {
  return ((GridLogView*) glv)->ShowLess(newsize);
}

void GridLogView::GetBodyRep(){
  if(!taMisc::iv_active) return;
  if(body != NULL) return;
  if(impl_ == NULL) impl_ = Get_IV_Impl();
  body = impl_->GetBodyRep();
  ivResource::ref(body);
}

void GridLogView::GetDataRep() {
  GridLogView_IV_Impl* glv_impl =(GridLogView_IV_Impl *) impl_;
  actual_range.min = view_range.min - log_mgr->data_range.min;
  actual_range.max = view_range.max - log_mgr->data_range.min;
  if(glv_impl->data_rep != NULL)
    ivResource::unref(glv_impl->data_rep);
  glv_impl->data_rep = 
    obj->Constr_Box(&actual_range,(int *)this, &GLVMore, &GLVLess);
  // if we are not scaling by anything, scale by all blocks
  ivResource::ref(glv_impl->data_rep);
  if(scale_on.size == 0) { // not scaling on any objects
    glv_impl->ScaleOnAllBlocks();
  }
  obj->hcsb->SetAdjustNotify(glv_impl->adjuster);
}
  


void GridLogView::InitDisplay(){
  if(!taMisc::iv_active) return;
  if(body == NULL) return; // loading
  impl_->InitDisplay();
}

void GridLogView::UpdateDisplay(TAPtr){
  if(!taMisc::iv_active) return;
  impl_->UpdateDisplay();
}

void GridLogView::UpdateAfterEdit(){
  LogView::UpdateAfterEdit();
  if(!taMisc::iv_active) return;
  if(body == NULL) return; // loading
  obj->hcsb->zero = scale_range.offset;
  obj->hcsb->SetRange(GetScaleRange());
  bool init = false;
  if((colorspec != NULL) && (obj->cs->spec != colorspec)) {
    taBase::SetPointer((TAPtr*)&(obj->cs->spec), colorspec);
    obj->cs->MapColors();
    init = true;
  }
  if(obj->fill_type != fill_type) {
    SetBlockFill(fill_type);
    init = true;
  }
  if( init == true)     InitDisplay();
  else UpdateDisplay();
}

void GridLogView::Initialize() {
  obj = NULL;
  view_bufsz = 3;
  view_shift = .2f;
  colorspec = NULL;
  block_size = 12;
  block_border_size = 2;
  fill_type = DT_GridViewSpec::COLOR;
}

void GridLogView::CutLinks() {
  LogView::CutLinks();
  scale_on.CutLinks();
  taBase::DelPointer((TAPtr*)&colorspec);
}

void GridLogView::Destroy() {
  CutLinks();
}

void GridLogView::SetBlockFill(DT_GridViewSpec::BlockFill b){
  fill_type = b;
  obj->SetBlockFill(b);
}

void GridLogView::UpdateGridLayout(DT_GridViewSpec::MatrixLayout ml){
   obj->dtgvs->UpdateLayout(ml);
   InitDisplay(); // gets new headerbox and grids
}


void GridLogView::ScriptScaleOn(char* nm, int on){
  if(on == true)
    taivMisc::RecordScript(GetPath() + ".scale_on.AddUnique(\"" + nm + "\");\n");
  else 
    taivMisc::RecordScript(GetPath() + ".scale_on.Remove(\"" + nm + "\");\n");
}
//////////////////////////
//    GraphLogView	//
//////////////////////////

#include <ta_misc/datagraph.h>

void GraphLogViewLabel::GetMasterViewer() {
  GraphLogView* glv = GET_MY_OWNER(GraphLogView);
  if((glv != NULL) && (glv->editor != NULL)) {
    viewer = glv->editor->viewer;
    master = glv->editor->graphg;
  }
}

void GraphLogView::Initialize() {
  x_axis_mode = AUTOSCALE;
  x_axis_index = 0;
  editor = NULL;
  labels.SetBaseType(&TA_GraphLogViewLabel); // make sure to use this type..
}

LogView_IV_Impl* GraphLogView::Get_IV_Impl(){ 
  return new GraphLogView_IV_Impl(this);
}


TypeDef* GraphLogView::DT_ViewSpecType() {
  return &TA_DT_GraphViewSpec;
}

TypeDef* GraphLogView::DA_ViewSpecType() {
  return &TA_DA_GraphViewSpec;
}

void GraphLogView::UpdateViewRange() {
  view_bufsz = log_mgr->data_bufsz; // graph doesn't have its own view (for now)
  view_shift = log_mgr->data_shift;
  view_range = log_mgr->data_range;
}
  
void GraphLogView::InitLinks() {
  taBase::Own(labels,this);
  LogView::InitLinks();
  UpdateViewRange();
}

void GraphLogView::Destroy() {
  CutLinks();
  if (editor) { delete editor; editor = NULL; }
}

void GraphLogView::CutLinks() {
  labels.CutLinks();
  LogView::CutLinks();
}

void GraphLogView::Copy_(const GraphLogView& cp) {
  x_axis_mode = cp.x_axis_mode;
  x_axis_index = cp.x_axis_index;
  labels = cp.labels;
}


void GraphLogView::UpdateAfterEdit() {
  LogView::UpdateAfterEdit();
  labels.SetBaseType(&TA_GraphLogViewLabel); // make sure to use this type..
  x_axis_index = MIN(log_mgr->data.size-1, x_axis_index);
  x_axis_index = MAX(0, x_axis_index);
  UpdateViewRange();
  if((editor != NULL) && (log_mgr != NULL) && 
     ((editor->x_axis_index != x_axis_index) || 
      (editor->x_axis_mode != (GraphEditor::XAxisMode) x_axis_mode)))
  {
    editor->x_axis_mode = (GraphEditor::XAxisMode)x_axis_mode;
    editor->SetXAxis(x_axis_index);
//    UpdateDisplay(); // called by above
  }
}

void GraphLogView::NewData() {
  if(!IsMapped() || !display_toggle) return;
  UpdateViewRange();
  if(editor != NULL) {
    editor->AddLastPoint();
  }
}

void GraphLogView::NewHead() {
  LogView::NewHead();
  if(!IsMapped() || !display_toggle) return;
  UpdateViewRange();
  if(editor != NULL) {
    editor->UpdateDisplay();
  }
}

int GraphLogView::SetXAxis(char* nm) {
  int dx = log_mgr->data.FindLeaf(nm);
  if(dx >= 0) {
    x_axis_index = dx;
    UpdateAfterEdit();
  }
  return x_axis_index;
}

void GraphLogView::Clear_impl() {
  view_range.min = view_range.max = 0; // collapse the view range..
//  UpdateDisplay();
  InitDisplay();
}

void GraphLogView::CloseWindow() {
  if(!taMisc::iv_active || (window == NULL)) return;
  ivResource::unref(body);	body = NULL;
  LogView::CloseWindow();
}

void GraphLogView::InitColors(){
  if(editor!=NULL) editor->InitColors();
}

void GLVSetXAxis(TAPtr tap, int idx){
  GraphLogView* glv = (GraphLogView *) tap;
  if(glv->x_axis_index != idx){
    glv->x_axis_index = idx;
    tabMisc::NotifyEdits(glv);
  }
}

void GraphLogView::GetBodyRep() {
  if(!taMisc::iv_active) return;
  editor = new GraphEditor((TAPtr) this, &(log_mgr->data),
			   (DT_GraphViewSpec*) viewspec, (ivWindow*) window);
  editor->Init();
  editor->SetOwnerXAxis = GLVSetXAxis;
  if(body != NULL) return;
  if(impl_ == NULL) impl_ = Get_IV_Impl();
  body = impl_->GetBodyRep();
  ivResource::ref(body);
};

void GraphLogView::UpdateFromBuffer() {
  if(!display_toggle)
    return;
  UpdateDisplay();
}

void GraphLogView::UpdateDisplay(TAPtr) {
  UpdateViewRange();
  if(editor != NULL) editor->UpdateDisplay();
}

void GraphLogView::InitDisplay() {
  UpdateViewRange();
  if(editor != NULL) editor->InitDisplay();
}

void GraphLogView::UpdateLineFeatures(){
  editor->dtvsp->UpdateLineFeatures();
  editor->InitDisplay();
}

void	GraphLogView::SetLineFeatures(DT_GraphViewSpec::ColorType	color_type,
					DT_GraphViewSpec::SequenceType sequence1,
					DT_GraphViewSpec::SequenceType sequence2,
					DT_GraphViewSpec::SequenceType sequence3,
					bool visible_only) {
  if(sequence3 == sequence1) sequence3 = DT_GraphViewSpec::NONE;
  if(sequence2 == sequence1) sequence2 = DT_GraphViewSpec::NONE;
  else if(sequence3 == sequence2) sequence3 = DT_GraphViewSpec::NONE;

  if((editor !=NULL) && (editor->dtvsp != NULL)){
    editor->dtvsp->color_type = color_type;
    editor->dtvsp->sequence_1 = sequence1;
    editor->dtvsp->sequence_2 = sequence2;
    editor->dtvsp->sequence_3 = sequence3;
    editor->dtvsp->UpdateLineFeatures(visible_only);
    editor->InitDisplay();
  }
    
}

void GraphLogView::SetBackground(RGBA background){
  if(editor == NULL) return;
  editor->dtvsp->background = background;
  editor->dtvsp->UpdateAfterEdit();
}
